#!/usr/bin/python
#transport_udp_session.py
#
#
#    Copyright DataHaven.NET LTD. of Anguilla, 2006
#    Use of this software constitutes acceptance of the Terms of Use
#      http://datahaven.net/terms_of_use.html
#    All rights reserved.
#
#


import os
import sys
import time
import struct
import cStringIO


from twisted.internet import reactor


import dhnio
import settings
import misc
import identitycache
import tmpfile 
import nameurl
import contacts

import automat

#------------------------------------------------------------------------------ 

DHN_STUN_SERVERS = set([
    ('67.207.147.183', 13478),
    ('208.78.96.185', 13478),
    ('178.49.192.122', 13478),
    ])

DHN_STUN_COMMANDS = { 'ROUTE':       'r:',
                     }

#------------------------------------------------------------------------------ 

CODES = {'p:':  'PING',
         'g:':  'GREETING',
         'd:':  'DATA',
         'r:':  'REPORT',
         'a:':  'ALIVE',}

COMMANDS = {'PING':         'p:',
            'GREETING':     'g:',
            'DATA':         'd:',
            'REPORT':       'r:',
            'ALIVE':        'a:',}

BLOCK_SIZE_LEVELS = {   0: 508,
                        1: 1460,
                        2: 2048,
                        3: 3072,
                        4: 4096,
                        5: 5120,
                        6: 6144,
                        7: 7168,
                        8: 8124,
                        }
#sz = 508
#BLOCK_SIZE_LEVELS = {   #0: 508,
#                        0: sz,
#                        1: sz,
#                        2: sz,
#                        3: sz,
#                        4: sz,
#                        5: sz,
#                        6: sz,
#                        7: sz,
#                        }
MAX_BLOCK_SIZE_LEVEL = len(BLOCK_SIZE_LEVELS) - 1

BLOCK_RETRIES = 999999
MAX_WINDOW_SIZE = 32
MIN_WINDOW_SIZE = 1
MAX_ACK_TIME_OUT = 4.0
MIN_ACK_TIME_OUT = 0.5
MAX_TIMEOUTS_RATIO = 0.5
FINE_TIMEOUTS_RATIO = 0.0

#------------------------------------------------------------------------------ 

_ParentObject = None

_IdToAddress = {}
_AddressToId = {}

_SendStatusCallback = None
_ReceiveStatusCallback = None
_SendControlFunc = None
_ReceiveControlFunc = None
_RegisterTransferFunc = None
_UnRegisterTransferFunc = None
_StateChangedCallbackFunc = None
_NetworkAddressChangedFunc = None

_TimerPing = None
_TimerAlive = None
_Timer30sec = None

_SendingDelay = 0.01
_SendingTask = None

_OutboxQueueDelay = 0.01
_OutboxQueueTask = None

#------------------------------------------------------------------------------ 

def init(parent):
    global _ParentObject
    dhnio.Dprint(4, 'transport_udp_session.init')
    _ParentObject = parent
    reactor.callLater(0, process_outbox_queue)
    reactor.callLater(1, process_sessions)
    StartTimers()

   
def shutdown():
    dhnio.Dprint(4, 'transport_udp_session.shutdown')
    StopTimers()
    
   
def A(address, event=None, arg=None):
    s = open_session(address)
    if event is not None:
        s.automat(event, arg)
    return s
      
      
def parent():
    global _ParentObject
    return _ParentObject

#------------------------------------------------------------------------------ 

def dhn_stun_servers_enabled():
    return True

def dhn_stun_servers_request(transport, datagram):
    for addr in DHN_STUN_SERVERS:
        try:
            transport.write(datagram, addr)
        except:
            pass

def dhn_stun_server_response(datagram, address):
    try:
        myipport, peeridurl, peeripport, deltatime = datagram.split(' ')
        if peeripport == '0':
            return
        myip, myport = myipport.split(':')
        myport = int(myport)
        peerip, peerport = peeripport.split(':')
        peerport = int(peerport)
    except:
        dhnio.DprintException()    
        return False
    myaddress = (myip, myport)
    peeraddress = (peerip, peerport)
    x, ip, port, x = misc.getLocalIdentity().getProtoParts('udp')
    # if ip and port and ip != myip and str(port) != str(myport):
        # network_address_changed(myaddress)
        # return
    mappedaddress = peeraddress
    if identitycache.HasLocalIP(peeridurl):
        mappedaddress = (identitycache.GetLocalIP(peeridurl), peeraddress[1]) 
    if mappedaddress != peeraddress:
        if not is_session_opened(mappedaddress):
            s = open_session(mappedaddress)
            dhnio.Dprint(6, 'transport_udp_session.dhn_stun_server_response made a new LOCAL session %s for user [%s], address=%s' % (s.name, nameurl.GetName(peeridurl), str(mappedaddress)))
            s.event('init', peeridurl)
            s.routed = True
    else:
        if not is_session_opened(peeraddress):
            s = open_session(peeraddress)
            dhnio.Dprint(6, 'transport_udp_session.dhn_stun_server_response made a new session %s for user [%s]' % (s.name, nameurl.GetName(peeridurl)))
            s.event('init', peeridurl)
            s.routed = True

def network_address_changed(newaddress):        
    global _NetworkAddressChangedFunc
    if _NetworkAddressChangedFunc is not None:
        _NetworkAddressChangedFunc(newaddress)

#------------------------------------------------------------------------------ 
    
class OutboxFile():
    def __init__(self, peer, remote_idurl, file_id, filename, filesize, description=''):
        global _RegisterTransferFunc
        self.peer = peer
        self.remote_idurl = remote_idurl
        self.file_id = file_id
        self.filename = filename
        # self.transfer_id = transfer_id
        self.description = description
        self.size = filesize
        self.blocks = {} 
        self.num_blocks = 0
        self.bytes_sent = 0
        self.bytes_out = 0
        self.block_id = -1
        self.blocks_timeouts = 0
        self.blocks_counter = 0
        # self.failed_blocks = {}
        # self.has_failed_blocks = False
        self.started = time.time()
        self.timeout = max( int(self.size/settings.SendingSpeedLimit()), 5)
        self.transfer_id = None
        if _RegisterTransferFunc is not None:
            self.transfer_id = _RegisterTransferFunc('send', 
                self.peer.remote_address, self.get_bytes_sent, self.filename, self.size, self.description)
            # _RegisterTransferFunc(callback=self.get_bytes_sent, transfer_id=self.transfer_id)
        # dhnio.Dprint(6, 'transport_udp_session.OutboxFile [%d] to %s' % (self.file_id, nameurl.GetName(self.remote_idurl)))

    def close(self):
        global _UnRegisterTransferFunc
        if _UnRegisterTransferFunc and self.transfer_id:
            _UnRegisterTransferFunc(self.transfer_id)
        # dhnio.Dprint(6, 'transport_udp_session.OutboxFile close [%d] to %s (%s)' % (self.file_id, nameurl.GetName(self.remote_idurl), str(self.peer)))
        # dt = time.time() - self.started 
        # print 'DONE', self.bytes_sent, self.bytes_out, dt, (self.bytes_sent/1024.0)/dt 

    def get_bytes_sent(self):
        return self.bytes_sent 

    def report_block(self, block_id):
        if not self.blocks.has_key(block_id):
            dhnio.Dprint(10, 'transport_udp_session.report_block WARNING unknown block_id in REPORT packet received from %s: [%d]' % (self.peer.remote_address, block_id))
            return
        self.bytes_sent += len(self.blocks[block_id])
        del self.blocks[block_id]
        # sys.stdout.write('%d <<<\n' % block_id)

    def send(self):
        do_send = False
        while len(self.peer.sliding_window) < self.peer.sliding_window_size and self.block_id < self.num_blocks and self.block_id >= 0:
            if not self.send_block(self.block_id):
                break
            self.peer.sliding_window[(self.file_id, self.block_id)] = time.time()
            self.block_id += 1
            do_send = True
        return do_send
    
    def check_blocks_timeouts(self):
        if self.blocks_counter < 10:
            return 0
        ratio = float(self.blocks_timeouts) / float(self.blocks_counter)
        self.blocks_counter = 0
        self.blocks_timeouts = 0
        return ratio
        
    def send_block(self, block_id):
        global _SendControlFunc
        if not self.blocks.has_key(block_id):
            dhnio.Dprint(8, 'transport_udp_session.send_block WARNING block_id=%d not found, file_id=%s, transfer_id=%s, blocks: %d' % (
                block_id, self.file_id, str(self.transfer_id), len(self.blocks) ))
            return False
        data = self.blocks[block_id]
        if _SendControlFunc is not None:
            more_bytes = _SendControlFunc(self.peer.last_bytes_sent, len(data))
            if more_bytes < len(data):
                return False
        self.peer.last_bytes_sent = len(data)
        datagram = COMMANDS['DATA']
        datagram += struct.pack('i', self.file_id)
        datagram += struct.pack('i', block_id)
        datagram += struct.pack('i', self.num_blocks)
        datagram += struct.pack('i', len(data))
        datagram += data
        self.bytes_out += len(data)
        # sys.stdout.write('>>> %d\n' % block_id)
        return self.peer.sendDatagram(datagram)

#    def mark_failed_block(self, block_id):
#        if not self.failed_blocks.has_key(block_id):
#            self.failed_blocks[block_id] = 0
#        self.failed_blocks[block_id] += 1
#        if self.failed_blocks[block_id] >= BLOCK_RETRIES:
#            self.has_failed_blocks = True
#            dhnio.Dprint(14, 'transport_udp_session.mark_failed_block (%d,%d) is failed to send, so transfer %s is failed' % (self.file_id, block_id, self.transfer_id))
#            return True
#        return False
    
    def read_blocks(self):
        fin = open(self.filename, 'rb')
        block_id = 0
        while True:
            block_data = fin.read(BLOCK_SIZE_LEVELS[self.peer.block_size_level])
            if block_data == '':
                break
            self.blocks[block_id] = block_data
            block_id += 1  
        fin.close()
        self.num_blocks = block_id
        self.block_id = 0     
        
    def is_timed_out(self):
        return time.time() - self.started > self.timeout

    def cancel(self):
        dhnio.Dprint(6, 'transport_udp_session.OutboxFile cancel timeout=%d' % self.timeout)
        self.timeout = 0

#------------------------------------------------------------------------------ 

class InboxFile():
    def __init__(self, peer, file_id, num_blocks):
        global _RegisterTransferFunc
        self.fd, self.filename = tmpfile.make("udp-in")
        self.peer = peer
        self.file_id = file_id
        self.num_blocks = num_blocks
        self.blocks = {}
        self.bytes_received = 0
        self.bytes_extra = 0
        self.transfer_id = None
        self.started = time.time()
        self.last_block_time = time.time()
        self.block_timeout = 0
        if _RegisterTransferFunc is not None:
            self.transfer_id = _RegisterTransferFunc('receive', self.peer.remote_address, self.get_bytes_received, self.filename, -1, '')
        # dhnio.Dprint(6, 'transport_udp_session.InboxFile {%s} [%d] from %s' % (self.transfer_id, self.file_id, str(self.peer.remote_address)))

    def close(self):
        global _UnRegisterTransferFunc
        if _UnRegisterTransferFunc is not None and self.transfer_id is not None:
            _UnRegisterTransferFunc(self.transfer_id)
        try:
            if self.bytes_received > 0 and self.bytes_extra > self.bytes_received * 0.1:
                dhnio.Dprint(10, 'transport_udp_session.InboxFile.close WARNING %s%% garbage traffic from %s' % (str(self.bytes_extra/float(self.bytes_received)), self.peer.remote_address))
        except:
            dhnio.DprintException()
        # dhnio.Dprint(6, 'transport_udp_session.InboxFile.close {%s} [%d] from %s' % (self.transfer_id, self.file_id, str(self.peer.remote_address)))
        # dt = time.time() - self.started 
        # print 'DONE', self.bytes_received, self.bytes_extra, dt, (self.bytes_received/1024.0)/dt 

    def get_bytes_received(self):
        return self.bytes_received

    def input_block(self, block_id, block_data):
        if block_id not in self.blocks:
            self.blocks[block_id] = block_data
            self.bytes_received += len(block_data)
        else:
            self.bytes_extra += len(block_data)
        self.last_block_time = time.time()
        self.block_timeout = max( int(len(block_data)/settings.SendingSpeedLimit()), 3) 
    
    def build(self):
        for block_id in xrange(len(self.blocks)):
            os.write(self.fd, self.blocks[block_id])
        os.close(self.fd)
        # dhnio.Dprint(10, 'transport_udp_server.InboxFile.build [%s] file_id=%d, blocks=%d' % (
        #     os.path.basename(filename), file_id, len(self.incommingFiles[file_id])))

    def is_timed_out(self):
        if self.block_timeout == 0:
            return False
        return time.time() - self.last_block_time > self.block_timeout

    
#------------------------------------------------------------------------------ 

class TransportUDPSession(automat.Automat):
    fast = True
    timers = {'timer-60sec':     (60,    ['PING',]),
              'timer-3min':      (60*10,  ['PING',]),
              }
    def __init__(self, address, name, state, debug_level):
        self.remote_address = address
        self.remote_idurl = None
        self.remote_name = None
        self.last_alive_packet_time = 0
        self.incommingFiles = {}
        self.outgoingFiles = {}
        self.maxOutgoingFiles = 1
        self.receivedFiles = {}
        self.outbox_queue = []
        self.tries = 0
        self.ack_timeout = MIN_ACK_TIME_OUT
        self.sliding_window = {}
        self.sliding_window_size = MIN_WINDOW_SIZE
        self.last_bytes_sent = 0
        # self.failedBlocks = {}
        self.hasFailedBlocks = False
        self.block_size_level = 0
        self.remote_id_request_last_time = 0
        self.routed = False
        automat.Automat.__init__(self, name, state, debug_level)
        
    def state_changed(self, oldstate, newstate):
        global _StateChangedCallbackFunc
        if _StateChangedCallbackFunc is not None:
            _StateChangedCallbackFunc(self.index, oldstate, newstate)
        
    def A(self, event, arg):
        #---CONNECTED---
        if self.state is 'CONNECTED':
            if event == 'datagram-received' :
                self.doReceiveData(arg)
            elif event == 'timer-30sec' and self.isAlive(arg) :
                self.doAlive(arg)
            elif event == 'shutdown' :
                self.state = 'CLOSED'
                self.doDestroyMe(arg)
            elif event == 'timer-30sec' and not self.isAlive(arg) :
                self.state = 'PING'
                self.doParentRequestRemoteID(arg)
                self.doPing(arg)
        #---AT_STARTUP---
        elif self.state is 'AT_STARTUP':
            if event == 'shutdown' :
                self.state = 'CLOSED'
                self.doDestroyMe(arg)
            elif event == 'init' :
                self.state = 'PING'
                self.doSaveIDURL(arg)
                self.doParentRequestRemoteID(arg)
                self.doPing(arg)
        #---PING---
        elif self.state is 'PING':
            if event == 'datagram-received' and self.isPingOrGreeting(arg) :
                self.state = 'GREETING'
                self.doReceiveData(arg)
                self.doGreeting(arg)
            elif event == 'datagram-received' and not self.isPingOrGreeting(arg) :
                self.doReceiveData(arg)
            elif event == 'shutdown' :
                self.state = 'CLOSED'
                self.doDestroyMe(arg)
            elif event == 'timer-60sec' :
                self.doParentRequestRemoteID(arg)
            elif event == 'timer-5sec' :
                self.doPing(arg)
            elif event == 'timer-3min' :
                self.state = 'CLOSED'
                self.doDestroyMe(arg)
                self.doParentRecreateMe(arg)
        #---GREETING---
        elif self.state is 'GREETING':
            if event == 'datagram-received' and self.isGreetingOrAlive(arg) :
                self.state = 'CONNECTED'
                self.doReceiveData(arg)
                self.doAlive(arg)
            elif event == 'datagram-received' and not self.isGreetingOrAlive(arg) :
                self.doReceiveData(arg)
            elif event == 'timer-5sec' :
                self.doGreeting(arg)
            elif event == 'shutdown' :
                self.state = 'CLOSED'
                self.doDestroyMe(arg)
            elif event == 'timer-30sec' :
                self.state = 'PING'
                self.doParentRequestRemoteID(arg)
                self.doPing(arg)
        #---CLOSED---
        elif self.state is 'CLOSED':
            pass

    def isPingOrGreeting(self, arg):
        try:
            datagram, addr = arg
        except:
            return False
        return datagram[:2] in [ COMMANDS['PING'], COMMANDS['GREETING'], ] 
            
    def isGreetingOrAlive(self, arg):
        return arg[0][:2] in [ COMMANDS['GREETING'], COMMANDS['ALIVE'], ]
    
    def isAlive(self, arg):
        return time.time() - self.last_alive_packet_time < 60.0 * 1.0 + 5.0
    
    def doSaveIDURL(self, arg):
        self.remote_idurl = arg
        self.remote_name = nameurl.GetName(self.remote_idurl)
  
    def doPing(self, arg):
        self.sendDatagram(COMMANDS['PING'])
        if not self.routed:
            if dhn_stun_servers_enabled():
                if self.remote_idurl:
                    request = DHN_STUN_COMMANDS['ROUTE']+('%s %s' % (misc.getLocalID(), self.remote_idurl))
                    dhn_stun_servers_request(parent().client.transport, request)
    
    def doGreeting(self, arg):
        self.sendDatagram(COMMANDS['GREETING']+misc.getLocalID())
    
    def doAlive(self, arg):
        self.sendDatagram(COMMANDS['ALIVE'])
    
    def doReceiveData(self, arg):
        global _ReceiveControlFunc
        try:
            datagram, addr = arg
        except:
            return
        io = cStringIO.StringIO(datagram)
        code = io.read(2)
        cmd = CODES.get(code, None)
        if cmd is None:
            # dhnio.Dprint(2, 'transport_udp_session.doReceiveData WARNING incorrect data from %s: [%s]' % (self.remote_address, code))
            return
        if _ReceiveControlFunc:
            seconds_pause = _ReceiveControlFunc(len(datagram))
            if seconds_pause > 0:
                # transport_udp.A().client.transport.pauseProducing()
                parent().client.client.transport.pauseProducing()
                # self.client.transport.pauseProducing()
                # reactor.callLater(seconds_pause, transport_udp.A().client.transport.resumeProducing)
                reactor.callLater(seconds_pause, parent().client.transport.resumeProducing)
        # dhnio.Dprint(8, '                                UDP:[%s] from %s' % (cmd, self.remote_address))
        self.last_alive_packet_time = time.time()
        #--- DATA
        if cmd == 'DATA':
            try:
                file_id = struct.unpack('i', io.read(4))[0]
                block_id = struct.unpack('i', io.read(4))[0]
                num_blocks = struct.unpack('i', io.read(4))[0]
                data_size = struct.unpack('i', io.read(4))[0]
            except:
                dhnio.DprintException()
                return
            if not self.incommingFiles.has_key(file_id):
                if self.receivedFiles.has_key(file_id):
                    if time.time() - self.receivedFiles[file_id] > 10:
                        # we receive the file, but 10 seconds since that we still receive the acks
                        # definitely something wrong - drop session 
                        self.automat('shutdown')
                        return 
                    self.sendDatagram(COMMANDS['REPORT'] + struct.pack('i', file_id) + struct.pack('i', block_id))
                    return
                if len(self.incommingFiles) >= 2 * self.maxOutgoingFiles:
                    # too many incoming files - drop session
                    self.automat('shutdown') 
                    return
                self.incommingFiles[file_id] = InboxFile(self, file_id, num_blocks)
            inp_data = io.read()
            if len(inp_data) != data_size:
                dhnio.Dprint(2, 'transport_udp_session.doReceiveData WARNING incorrect datagram received from %s, not full data' % str(self.remote_address))
                return
            self.incommingFiles[file_id].input_block(block_id, inp_data) 
            self.sendDatagram(COMMANDS['REPORT'] + struct.pack('i', file_id) + struct.pack('i', block_id))
            # dhnio.Dprint(12, 'transport_udp_session.doReceiveData [%d] (%d/%d) from %s, state=%s' % (file_id, block_id, num_blocks, self.remote_address, self.state))
            if len(self.incommingFiles[file_id].blocks) == num_blocks:
                # dhnio.Dprint(12, 'transport_udp_session.doReceiveData new file [%d] (%d/%d) from %s' % (file_id, block_id, num_blocks, self.remote_address))
                self.incommingFiles[file_id].build()
                file_received(self.incommingFiles[file_id].filename, 'finished', 'udp', self.remote_address)
                self.incommingFiles[file_id].close()   
                del self.incommingFiles[file_id]   
                self.receivedFiles[file_id] = time.time()
                self.eraseOldFileIDs()
        #--- REPORT
        elif cmd == 'REPORT':
            try:
                file_id = struct.unpack('i', io.read(4))[0]
            except:
                dhnio.DprintException()
                return
            if not self.outgoingFiles.has_key(file_id):
                # if not self.receivedFiles.has_key(file_id):
                #     dhnio.Dprint(8, 'transport_udp_session.doReceiveData WARNING unknown file_id in REPORT packet received from %s: [%d]' % (self.remote_address, file_id))
                return
            try:
                block_id = struct.unpack('i', io.read(4))[0]
            except:
                dhnio.DprintException()
                return
            self.outgoingFiles[file_id].report_block(block_id)
            self.sliding_window.pop((file_id, block_id), None)
            self.sliding_window_size += 1
            if self.sliding_window_size > MAX_WINDOW_SIZE:
                self.sliding_window_size = MAX_WINDOW_SIZE
            if len(self.outgoingFiles[file_id].blocks) == 0:
                self.reportSentDone(file_id)
                self.closeOutboxFile(file_id)
            # dhnio.Dprint(8, 'transport_udp_session.doReceiveData REPORT on [%d,%d] received, blocks=%d, window=%d' % (file_id, block_id, len(self.outgoingFiles[file_id]), self.sliding_window_size))
        #--- GREETING
        elif cmd == 'GREETING':
            greeting_idurl = io.read()
            dhnio.Dprint(12, 'transport_udp_session.doReceiveData GREETING "%s" from %s, remote_idurl=%s, remote_address=%s, state=%s' % (
                greeting_idurl[:120], str(addr), self.remote_idurl, self.remote_address, self.state))
            if greeting_idurl:
                if nameurl.UrlMake(parts=nameurl.UrlParse(greeting_idurl)) != greeting_idurl:
                    dhnio.Dprint(2, 'transport_udp_session.doReceiveData WARNING incorrect idurl=%s in GREETING packet from %s' % (greeting_idurl, self.name))
                    self.sendDatagram(COMMANDS['PING'])
                else:
                    if self.remote_idurl is None:
                        self.remote_idurl = greeting_idurl
                        self.remote_name = nameurl.GetName(self.remote_idurl)
                        dhnio.Dprint(6, 'transport_udp_session.doReceiveData got idurl=%s in GREETING packet from %s' % (greeting_idurl, self.name))
                        parent().automat('recognize-remote-id', (self.index, self.remote_idurl))
                    else:
                        if self.remote_idurl != greeting_idurl:
                            dhnio.Dprint(2, 'transport_udp_session.doReceiveData WARNING wrong idurl=%s in GREETING packet from %s' % (greeting_idurl, self.name))
                    ident = contacts.getContact(greeting_idurl)
                    if ident:
                        udpip = ident.getIP('udp')
                        if udpip:
                            if udpip != addr[0]:
                                parent().automat('detect-remote-ip', (self.index, self.remote_idurl))
                    self.sendDatagram(COMMANDS['ALIVE'])
            else:
                dhnio.Dprint(2, 'transport_udp_session.doReceiveData WARNING did not found idurl in GREETING packet from %s, send PING again' % self.name)
                self.sendDatagram(COMMANDS['PING'])
        #--- PING
        elif cmd == 'PING':
            self.sendDatagram(COMMANDS['GREETING']+misc.getLocalID())
        #--- ALIVE            
        elif cmd == 'ALIVE':
            pass

    def doDestroyMe(self, arg):
        # dhnio.Dprint(6, 'transport_udp_session.doDestroyMe %s' % self.name)
        file_ids_to_remove = self.incommingFiles.keys()
        for file_id in file_ids_to_remove:
            try:
                os.close(self.incommingFiles[file_id].fd)
            except:
                dhnio.DprintException()
            file_received(self.incommingFiles[file_id].filename, 'failed', 'udp', self.remote_address, None, 'session has been closed')
            self.incommingFiles[file_id].close()   
            del self.incommingFiles[file_id]
        file_ids_to_remove = self.outgoingFiles.keys()
        for file_id in file_ids_to_remove:
            file_sent(self.remote_address, self.outgoingFiles[file_id].filename, 'failed', 'udp', None, 'session has been closed')
            self.closeOutboxFile(file_id)
        self.receivedFiles.clear()
        self.sliding_window.clear()
        automat.objects().pop(self.index)
        
    def doParentRequestRemoteID(self, arg):
        if self.remote_idurl:
            # dhnio.Dprint(8, 'transport_udp_session.doParentRequestRemoteID [%s] for session %s' % (self.remote_name, self.name))
            parent().automat('child-request-remote-id', self.remote_idurl)

    def doParentRecreateMe(self, arg):
        if self.remote_idurl:
            parent().automat('child-request-recreate', self.remote_idurl)


    def sendDatagram(self, data):
        # dhnio.Dprint(10, '    [UDP] %d bytes to %s (%s)' % (len(data), self.remote_address, nameurl.GetName(self.remote_idurl)))
        if parent().client is not None:
        # if transport_udp.A().client is not None:
            try:
                # transport_udp.A().client.transport.write(data, self.remote_address)
                parent().client.transport.write(data, self.remote_address)
                return True
            except:
                return False 
        else:
            dhnio.Dprint(2, 'transport_udp_session.sendDatagram WARNING client is None')
            return False      
        
    def makeFileID(self):
        return int(str(int(time.time() * 100.0))[4:])   

#    def register_outbox_file(self, filename, sz, callback, description=''): 
#        global _RegisterTransferFunc
#        if _RegisterTransferFunc is None:
#            return None
#        transfer_id = _RegisterTransferFunc('send', self.remote_address, callback, filename, sz, description)
#        return transfer_id
   
    def addOutboxFile(self, filename, description=''):
        self.outbox_queue.append((filename, description))
        # transfer_id, filesize = self.register_outbox_file(filename, description)
        # if transfer_id:
        #     self.outbox_queue.append((filename, filesize, transfer_id, description))
        # else:
        #     file_sent(self.remote_address, filename, 'failed', 'udp', None, 'unable to register transfer')
        # dhnio.Dprint(6, 'transport_udp_session.addOutboxFile [%s], total outbox files: %d' % (os.path.basename(filename), len(self.outbox_queue)))
        
    def putOutboxFile(self, filename, description=''):
        self.outbox_queue.insert(0, (filename, description))
        # transfer_id, filesize = self.register_outbox_file(filename, description)
        # if transfer_id:
        #     self.outbox_queue.insert(0, (filename, filesize, transfer_id, description)) 
        # else:
        #     file_sent(self.remote_address, filename, 'failed', 'udp', None, 'unable to register transfer')
        # dhnio.Dprint(6, 'transport_udp_session.putOutboxFile [%s], total outbox files: %d' % (os.path.basename(filename), len(self.outbox_queue)))
    
    def eraseOldFileIDs(self):
        if len(self.receivedFiles) > 10:
            file_ids = self.receivedFiles.keys()
            cur_tm = time.time()
            for file_id in file_ids:
                if cur_tm - self.receivedFiles[file_id] > 60 * 20:
                    del self.receivedFiles[file_id]
            del file_ids 

    def closeOutboxFile(self, file_id):
        self.outgoingFiles[file_id].close()
        del self.outgoingFiles[file_id]

    def reportSentDone(self, file_id):
        file_sent(self.remote_address, self.outgoingFiles[file_id].filename, 'finished', 'udp')
        self.block_size_level += 1
        if self.block_size_level > MAX_BLOCK_SIZE_LEVEL:
            self.block_size_level = MAX_BLOCK_SIZE_LEVEL
        # print self.block_size_level
    
    def reportSentFailed(self, file_id, why=''):
        file_sent(self.remote_address, self.outgoingFiles[file_id].filename, 'failed', 'udp', None, why)
        self.block_size_level -= 2
        if self.block_size_level < 0:
            self.block_size_level = 0
        # print self.block_size_level

    def processOutboxQueue(self):
        # global _UnRegisterTransferFunc
        has_reads = False
        while len(self.outbox_queue) > 0 and len(self.outgoingFiles) < self.maxOutgoingFiles:        
            # filename, filesize, transfer_id, description = self.outbox_queue.pop(0)
            filename, description = self.outbox_queue.pop(0)
            has_reads = True
            # we have a queue of files to be sent
            # somehow file may be removed before we start sending it
            # so we check it here and skip not existed files
            if not os.path.isfile(filename):
                file_sent(self.remote_address, filename, 'failed', 'udp', None, 'file were removed')
                # if _UnRegisterTransferFunc:
                #     _UnRegisterTransferFunc(transfer_id)
                continue
            try:
                sz = os.path.getsize(filename)
            except:
                file_sent(self.remote_address, filename, 'failed', 'udp', None, 'can not get file size')
                continue
#            if os.access(filename, os.R_OK):
#                file_sent(self.remote_address, filename, 'failed', 'udp', None, 'access denied')
#                continue
            file_id = self.makeFileID()
            f = OutboxFile(self, self.remote_idurl, file_id, filename, sz, description)
            f.read_blocks()
            self.outgoingFiles[file_id] = f
        return has_reads

    def checkIncomingFiles(self):
        for file_id in self.incommingFiles.keys():
            if self.incommingFiles[file_id].is_timed_out():
                self.automat('shutdown')
                return

    def processSending(self):
        has_sends = False
        has_timedout = False
        failed_ids = []
        for file_id in self.outgoingFiles.keys():
            if self.outgoingFiles[file_id].send():
                has_sends = True
            if self.outgoingFiles[file_id].is_timed_out():
                if self.outgoingFiles[file_id].timeout == 0:
                    failed_ids.append((file_id, 'canceled'))
                else:
                    # failed_ids.append((file_id, 'timeout'))
                    has_timedout = True
        for file_id, why in failed_ids:
            self.reportSentFailed(file_id, why)
            self.closeOutboxFile(file_id)
        del failed_ids
        if has_timedout:
            # if some packets we currently sending is timed out
            # all other in the outbox will fail too - so erase all.
            self.clearOutboxQueue()
        has_sends = has_sends or self.checkWindow()
        return has_sends
    
    def clearOutboxQueue(self):
        failed = []
        while len(self.outbox_queue) > 0:
            filename, filesize, transfer_id, description = self.outbox_queue.pop(0)
            failed.append((filename, filesize, transfer_id, description))
        for filename, filesize, transfer_id, description in failed:
            reactor.callLater(0, file_sent, self.remote_address, filename, 'failed', 'udp', None, 'timeout')
        del failed
            
    def checkWindow(self):
        if len(self.sliding_window) == 0:
            return False
        to_remove = [] 
        to_update = {}
        cur_tm = time.time()
        do_send = True
        for block_info, tm in self.sliding_window.items():
            if cur_tm - tm < self.ack_timeout: # give some time to get an Ack. 
                continue
            to_remove.append(block_info)
            file_id = block_info[0]
            block_id = block_info[1]
            if not self.outgoingFiles.has_key(file_id):
                # dhnio.Dprint(2, 'transport_udp_session.checkWindow WARNING unknown file_id [%d] from %s' % (file_id, nameurl.GetName(self.remote_idurl)))
                continue
            if not self.outgoingFiles[file_id].blocks.has_key(block_id):
                dhnio.Dprint(2, 'transport_udp_session.checkWindow WARNING unknown block_id: %d' % block_id)
                continue
            do_send = True
            to_update[block_info] = cur_tm
            outFile = self.outgoingFiles[file_id]
            outFile.blocks_timeouts += 1
            outFile.send_block(block_id)
            timeouts_ratio = outFile.check_blocks_timeouts()
            self.adjust_ack_timeout(timeouts_ratio)
#            self.outgoingFiles[file_id].send_block(block_id)
#            isFailed = self.outgoingFiles[file_id].mark_failed_block(block_id)
#            if not isFailed:
#                to_update[block_info] = cur_tm
        for block_info in to_remove:
            del self.sliding_window[block_info]
            # dhnio.Dprint(8, 'transport_udp_session.doCheckWindow (%d,%d) removed, window=%d' % (file_id, block_id, len(self.sliding_window)))
        for block_info, tm in to_update.items():
            self.sliding_window[block_info] = tm
            # dhnio.Dprint(8, 'transport_udp_session.doCheckWindow (%s) resending to %s, window=%d' % (block_info, self.remote_address, len(self.sliding_window)))
        if len(to_remove) > 0:
            self.sliding_window_size = int( self.sliding_window_size / 2.0 )
            if self.sliding_window_size < MIN_WINDOW_SIZE:
                self.sliding_window_size = MIN_WINDOW_SIZE
            # dhnio.Dprint(8, 'transport_udp_session.doCheckWindow decreased, window=%d' % self.sliding_window_size)
        del to_remove
        del to_update
        return 

    def adjust_ack_timeout(self, timeouts_ratio):
        if timeouts_ratio >= MAX_TIMEOUTS_RATIO:
            self.ack_timeout *= 2.0
            if self.ack_timeout > MAX_ACK_TIME_OUT:
                self.ack_timeout = MAX_ACK_TIME_OUT
                # TODO transfer is failed - udp packets did not get acket
                # drop connection!
                # TODO change transport_udp_server to drop connection
                self.automat('shutdown')
            # else:
            #     print('adjust_ack_timeout %f' % self.ack_timeout)
        elif timeouts_ratio <= FINE_TIMEOUTS_RATIO:
            self.ack_timeout /= 2.0
            if self.ack_timeout < MIN_ACK_TIME_OUT:
                self.ack_timeout = MIN_ACK_TIME_OUT
                # TODO transfer is great!
                #      increase window size, block size? 
            # else:
            #     print('adjust_ack_timeout %f' % self.ack_timeout)

    def cancel(self, transferID):
        for i in xrange(len(self.outbox_queue)):
            filename, filesize, transfer_id, description = self.outbox_queue[i]
            if transferID == transfer_id:
                del self.outbox_queue[i]
                break
        for file_id in self.outgoingFiles.keys():
            if self.outgoingFiles[file_id].transfer_id == transferID:
                self.outgoingFiles[file_id].cancel()
                # self.receivedFiles[file_id] = time.time()
                break
        for file_id in self.incommingFiles.keys():
            if self.incommingFiles[file_id].transfer_id == transferID:
                self.incommingFiles[file_id].build()
                file_received(self.incommingFiles[file_id].filename, 'failed', 'udp', self.remote_address)              
                self.incommingFiles[file_id].close()
                del self.incommingFiles[file_id]
                self.receivedFiles[file_id] = time.time()
                break

#------------------------------------------------------------------------------ 
    
def check_outbox_queue():    
    has_reads = False
    for sess in sessions():
        if sess.processOutboxQueue():
            has_reads = True
    if has_reads:
        check_sending_queue()
    return has_reads 
    
def check_sending_queue():    
    has_sends = False
    for sess in sessions():
        if sess.processSending():
            has_sends = True
    return has_sends

def check_incoming_files():
    for sess in sessions():
        sess.checkIncomingFiles()
    
def process_outbox_queue():
    global _OutboxQueueDelay
    global _OutboxQueueTask
    
    has_reads = check_outbox_queue()

    _OutboxQueueDelay = misc.LoopAttenuation(
        _OutboxQueueDelay, 
        has_reads, 
        settings.MinimumSendingDelay(), 
        settings.MaximumSendingDelay(),)
    
    # attenuation
    _OutboxQueueTask = reactor.callLater(_OutboxQueueDelay, process_outbox_queue)
        

def process_sessions():
    global _SendingDelay
    global _SendingTask
    
    check_incoming_files()
    
    has_sends = check_sending_queue()
    
    _SendingDelay = misc.LoopAttenuation(
        _SendingDelay, 
        has_sends, 
        settings.MinimumSendingDelay(), 
        settings.MaximumSendingDelay(),)
    
    # attenuation
    _SendingTask = reactor.callLater(_SendingDelay, process_sessions)
    
#------------------------------------------------------------------------------ 

def sessions():
    for a in automat.objects().values():
        if a.name.startswith('udp_'):
            yield a


def open_session(address):
    name = 'udp_%s_%s' % (address[0].replace('.','_'), str(address[1]))
    for a in automat.objects().values():
        if a.name == name:
            return a
    a = TransportUDPSession(address, name, 'AT_STARTUP', 12)
    return a


def get_session(address):
    name = 'udp_%s_%s' % (address[0].replace('.','_'), str(address[1]))
    for a in automat.objects().values():
        if a.name == name:
            return a
    return None


def get_session_by_index(index):
    return automat.objects().get(index, None)


def is_session_opened(address):
    name = 'udp_%s_%s' % (address[0].replace('.','_'), str(address[1]))
    for a in automat.objects().values():
        if a.name == name:
            return True
    return False
        

def outbox_file(address, filename, fast=False, description=''):
#    dhnio.Dprint(6, 'transport_udp_session.outbox_file %s to %s (%s)' % (
#        os.path.basename(filename), nameurl.GetName(idurl), str(address)))
    s = open_session(address)
    if s.state != 'CONNECTED':
        file_sent(s.remote_address, filename, 'failed', 'udp', None, 'remote peer is not connected')
        return
    if fast:
        s.putOutboxFile(filename, description)
    else:
        s.addOutboxFile(filename, description)
    check_outbox_queue()


def cancel(transferID):
    for sess in sessions():
        sess.cancel(transferID)
    

def big_event(event, arg=None):
    for sess in sessions():
        sess.event(event, arg)


def data_received(datagram, address):
    try:
        from transport_control import black_IPs_dict
        if address[0] in black_IPs_dict().keys() or '.'.join(address[0].split('.')[:-1]) in black_IPs_dict().keys():
            # dhnio.Dprint(12, 'transport_udp_session.data_received %d bytes from BLACK IP: %s, %s' % (len(datagram), str(address[0]), str(black_IPs_dict()[address[0]])) )
            return
    except:
        pass
    if dhn_stun_servers_enabled():
        if address in DHN_STUN_SERVERS:
            dhn_stun_server_response(datagram, address)
            return
    if not is_session_opened(address):
        s = open_session(address)
        idurl = identitycache.GetIDURLByIPPort(address[0], address[1])
        if idurl:
            dhnio.Dprint(6, 'transport_udp_session.data_received from KNOWN user, made a new session %s for [%s]' % (s.name, nameurl.GetName(idurl)))
            s.event('init', idurl)
        else:
            dhnio.Dprint(6, 'transport_udp_session.data_received from UNKNOWN user, made a new session %s' % s.name)
            s.event('init', None)
    A(address, 'datagram-received', (datagram, address))


def shutdown_all_sessions():
    big_event('shutdown')
    
    
def file_received(filename, status, proto='', host=None, error=None, message=''):
    global _ReceiveStatusCallback
    if _ReceiveStatusCallback is not None:
        _ReceiveStatusCallback(filename, status, proto, host, error, message)
    
    
def file_sent(host, filename, status, proto='', error=None, message=''):
    global _SendStatusCallback
    if _SendStatusCallback is not None:
        _SendStatusCallback(host, filename, status, proto, error, message)


#def timer30sec():
#    global _Timer30sec
#    big_event('timer-30sec')
#    _Timer30sec = reactor.callLater(ndom.randint(25,35), timer30sec)


def timerAlive():
    global _TimerAlive
    big_event('timer-30sec')
    _TimerAlive = reactor.callLater(30, timerAlive)


def timerPing():
    global _TimerPing
    big_event('timer-5sec')
    _TimerPing = reactor.callLater(5, timerPing)
    
#------------------------------------------------------------------------------ 
    
def SetReceiveStatusCallback(cb):
    global _ReceiveStatusCallback
    _ReceiveStatusCallback = cb
    
def SetSendStatusCallback(cb):
    global _SendStatusCallback
    _SendStatusCallback = cb
    
def SetSendControlFunc(f):
    global _SendControlFunc
    _SendControlFunc = f
    
def SetReceiveControlFunc(f):
    global _ReceiveControlFunc
    _ReceiveControlFunc = f

def SetRegisterTransferFunc(f):
    global _RegisterTransferFunc
    _RegisterTransferFunc = f

def SetUnRegisterTransferFunc(f):
    global _UnRegisterTransferFunc
    _UnRegisterTransferFunc = f

def SetStateChangedCallbackFunc(f):
    global _StateChangedCallbackFunc
    _StateChangedCallbackFunc = f
    
def SetNetworkAddressChangedFunc(f):
    global _NetworkAddressChangedFunc
    _NetworkAddressChangedFunc = f
    
#------------------------------------------------------------------------------ 

    
def StartTimers():
    global _Timer30sec
    global _TimerAlive
    global _TimerPing
#    if _Timer30sec is None:
#        timer30sec()
    if _TimerAlive is None: 
        timerAlive()
    if _TimerPing is None:
        timerPing()


def StopTimers():
    global _Timer30sec
    global _TimerAlive
    global _TimerPing
#    if _Timer30sec:
#        _Timer30sec.cancel()
#        _Timer30sec = None
    if _TimerAlive:
        _TimerAlive.cancel()
        _TimerAlive = None
    if _TimerPing:
        _TimerPing.cancel()
        _TimerPing = None








#!/usr/bin/python
#transport_tcp.py
#
#
#    Copyright DataHaven.NET LTD. of Anguilla, 2006
#    Use of this software constitutes acceptance of the Terms of Use
#      http://datahaven.net/terms_of_use.html
#    All rights reserved.
#
#


import os
import sys
import time

if __name__ == '__main__':
    sys.path.insert(0, os.path.abspath('..'))

try:
    from twisted.internet import reactor
except:
    sys.exit('Error initializing twisted.internet.reactor in transport_tcp.py')

from twisted.internet import task
from twisted.internet import protocol
from twisted.protocols import basic
from twisted.internet.defer import Deferred 
from twisted.internet.error import ConnectionDone

import dhnio
import tmpfile

#------------------------------------------------------------------------------ 

_SendStatusFunc = None
_ReceiveStatusFunc = None
_SendControlFunc = None
_ReceiveControlFunc = None
_RegisterTransferFunc = None
_UnRegisterTransferFunc = None
_ByTransferID = {}
_OpenedConnections = {}
_ConnectionsCounter = 0
_StartedConnections = {}
_SingleConnectionHosts = set()
_InternalPort = 7771
_RedirectionsMap = {}

#------------------------------------------------------------------------------

def init(   sendStatusFunc=None, 
            receiveStatusFunc=None, 
            sendControl=None, 
            receiveControl=None,
            registerTransferFunc=None,
            unregisterTransferFunc=None):
    dhnio.Dprint(4, 'transport_tcp.init')
    global _SendStatusFunc
    global _ReceiveStatusFunc
    global _RegisterTransferFunc
    global _UnRegisterTransferFunc
    if sendStatusFunc is not None:
        _SendStatusFunc = sendStatusFunc
    if receiveStatusFunc is not None:
        _ReceiveStatusFunc = receiveStatusFunc
    if sendControl is not None:
        _SendControlFunc = sendControl 
    if registerTransferFunc is not None:
        _RegisterTransferFunc = registerTransferFunc
    if unregisterTransferFunc is not None:
        _UnRegisterTransferFunc = unregisterTransferFunc
    

def sendsingle(filename, host, port, do_status_report=True, result_defer=None, description=''):
    sender = SingleSendingFactory(
        filename, host, int(port), 
        result_defer=result_defer, 
        do_status_report=do_status_report, 
        send_control_func=None, # (send_control_func or _SendControlFunc),
        description=description)
    reactor.connectTCP(host, int(port), sender)


def send(filename, host, port, do_status_report=True, result_defer=None, description=''):
    global _OpenedConnections
    global _StartedConnections
    global _RedirectionsMap
    remoteaddress = (host, int(port))
#    redirected = _RedirectionsMap.get(remoteaddress, None)
#    if redirected and redirected in _OpenedConnections:
#        if len(_OpenedConnections[redirected]) > 0: 
#            _OpenedConnections[redirected][0].sendfile(filename, do_status_report, result_defer, description)
#            return
    if remoteaddress in _StartedConnections:
        _StartedConnections[remoteaddress].addoutboxfile(filename, do_status_report, result_defer, description)
        return
    if remoteaddress in _OpenedConnections:
        if len(_OpenedConnections[remoteaddress]) > 0: 
            _OpenedConnections[remoteaddress][0].sendfile(filename, do_status_report, result_defer, description)
            return
    conn = TCPFactory()
    conn.addoutboxfile(filename, do_status_report, result_defer, description)
    _StartedConnections[remoteaddress] = conn
    reactor.connectTCP(host, int(port), conn)
    # dhnio.Dprint(14, '    %s' % str(_OpenedConnections))
    # dhnio.Dprint(14, '    %s' % str(_StartedConnections))


def cancel(transferID):
    global _ByTransferID
    inst = _ByTransferID.get(transferID, None)
    if not inst:
        dhnio.Dprint(2, "transport_tcp.cancel ERROR %s is not found" % str(transferID))
        return False
    if isinstance(inst, SingleSendingProtocol):
        inst.fileSender.stopProducing()
        dhnio.Dprint(8, "transport_tcp.cancel sending %s" % str(transferID))
        return True
#    if isinstance(inst, SingleReceiveProtocol):
#        reactor.callLater(0, inst.transport.loseConnection)
#        dhnio.Dprint(8, "transport_tcp.cancel receiving %s" % str(transferID))
#        return True
    if isinstance(inst, TCPProtocol):
        # reactor.callLater(0, inst.transport.loseConnection)
        inst.disconnect()
        # inst.canceltransfer(transferID)
        dhnio.Dprint(8, "transport_tcp.cancel transfer %s - disconnect channel" % str(transferID))
        return True
    dhnio.Dprint(2, "transport_tcp.cancel ERROR incorrect instance for %s" % str(transferID))
    return False
        

def receive(port, receive_control_func=None):
    global _ReceiveControlFunc
    global _InternalPort
    dhnio.Dprint(8, "transport_tcp.receive going to listen on port "+ str(port))

    def _try_receiving(port, count, receive_control_func):
        global _InternalPort
        dhnio.Dprint(8, "transport_tcp.receive count=%d" % count)
        # f = SingleReceiveFactory(receive_control_func)
        f = TCPFactory()
        try:
            mylistener = reactor.listenTCP(int(port), f)
            _InternalPort = int(port)
        except:
            mylistener = None
            dhnio.DprintException()
        return mylistener

    def _loop(port, result, count, receive_control_func):
        if count > 3:
            dhnio.Dprint(1, "transport_tcp.receive WARNING port %s is busy!" % str(port))
            result.callback(None)
            return
        l = _try_receiving(port, count, receive_control_func)
        if l is not None:
            dhnio.Dprint(8, "transport_tcp.receive started on port "+ str(port))
            result.callback(l)
            return
        reactor.callLater(1, _loop, port, result, count+1, receive_control_func)

    res = Deferred()
    _loop(port, res, 0, receive_control_func or _ReceiveControlFunc)
    return res


def disconnect(host, port):
    global _OpenedConnections
    connections = _OpenedConnections.get((str(host), int(port)), [])
    dhnio.Dprint(10, 'transport_tcp.disconnect %s %d, connections=%s' % (host, int(port), connections))
    for connection in connections:
        connection.disconnect()


def disconnect_all():
    global _OpenedConnections
    for peer, connections in _OpenedConnections.items():
        for connection in connections:
            connection.disconnect()
    
#------------------------------------------------------------------------------ 

def opened_connections():
    global _OpenedConnections
    return _OpenedConnections

def opened_connections_count():
    global _ConnectionsCounter
    return _ConnectionsCounter

def started_connections():
    global _StartedConnections
    return _StartedConnections

def single_connection_hosts():
    global _SingleConnectionHosts
    return _SingleConnectionHosts

#------------------------------------------------------------------------------

class SingleFileSender(basic.FileSender):
    
    def __init__(self, protocol):
        self.protocol = protocol
        self.bytesLastRead = 0
    
    def resumeProducing(self):
        #sys.stdout.write('=')
        chunk = ''
        if self.file:
            more_bytes = self.CHUNK_SIZE
            if self.protocol.factory.send_control_func:
                more_bytes = self.protocol.factory.send_control_func(self.bytesLastRead, self.CHUNK_SIZE)
            try:
                chunk = self.file.read(more_bytes)
            except:
                chunk = ''
            self.bytesLastRead = len(chunk)
        if not chunk:
            self.file = None
            self.bytesLastRead = 0
            self.consumer.unregisterProducer()
            if self.deferred:
                self.deferred.callback(self.lastSent)
                self.deferred = None
            return
        if self.transform:
            chunk = self.transform(chunk)
        self.consumer.write(chunk)
        self.lastSent = chunk[-1]


## In Putter "self.factory" references the parent object, so we can
## access arguments like "host", "port", and "filename"
class SingleSendingProtocol(protocol.Protocol):
    fin = None
    sentBytes = 0
    transfer_id = None
    
    def getSentBytes(self):
        return self.sentBytes
    
    def connectionMade(self):
        global _SendStatusFunc
        global _RegisterTransferFunc
        global _ByTransferID
        # dhnio.Dprint(14, 'transport_tcp.SingleSendingProtocol.connectionMade with %s' % str(self.transport.getPeer()))
        if not os.path.isfile(self.factory.filename):
            # dhnio.Dprint(6, 'transport_tcp.SingleSendingProtocol.connectionMade WARNING file %s was not found but we get connected to the %s' % (self.factory.filename, self.factory.host))
            if self.factory.do_status_report:
                _SendStatusFunc(self.transport.getPeer(), self.factory.filename,
                    'failed', 'tcp', None, 'file was not found')
            self.transport.loseConnection()
            if self.factory.result_defer is not None:
                if not self.factory.result_defer.called:
                    self.factory.result_defer.errback(Exception('failed'))
                    self.factory.result_defer = None
            return
        try:
            sz = os.path.getsize(self.factory.filename)
            self.fin = open(self.factory.filename, 'rb')
        except:
            dhnio.DprintException()
            if self.factory.do_status_report:
                _SendStatusFunc(self.transport.getPeer(), self.factory.filename,
                    'failed', 'tcp', None, 'error opening file')
            self.transport.loseConnection()
            if self.factory.result_defer is not None:
                if not self.factory.result_defer.called:
                    self.factory.result_defer.errback(Exception('failed'))
                    self.factory.result_defer = None
            return

        self.fileSender = SingleFileSender(self)
        d = self.fileSender.beginFileTransfer(self.fin, self.transport, self.transformData)
        d.addCallback(self.finishedTransfer)
        d.addErrback(self.transferFailed)
        self.transfer_id = _RegisterTransferFunc(
            'send', (self.transport.getPeer().host, int(self.transport.getPeer().port)), 
            self.getSentBytes, self.factory.filename, sz, self.factory.description)
        _ByTransferID[self.transfer_id] = self

    def transformData(self, data):
        self.sentBytes += len(data)
        return data

    def finishedTransfer(self, result):
        global _SendStatusFunc
        self.transport.loseConnection()
        try:
            self.fin.close()
        except:
            dhnio.Dprint(1, 'transport_tcp.SingleSendingProtocol.finishedTransfer ERROR close file failed')
            dhnio.DprintException()
        if self.factory.do_status_report:
            _SendStatusFunc(self.transport.getPeer(), self.factory.filename, 'finished', 'tcp')
        if self.factory.result_defer is not None:
            if not self.factory.result_defer.called:
                self.factory.result_defer.callback('finished')
                self.factory.result_defer = None
        # dhnio.Dprint(14, 'transport_tcp.SingleSendingProtocol.finishedTransfer host=%s' % (str(self.factory.host)))

    def transferFailed(self, err):
        global _SendStatusFunc
        self.transport.loseConnection()
        try:
            self.fin.close()
        except:
            dhnio.Dprint (1, 'transport_tcp.SingleSendingProtocol.transferFailed ERROR closing file %s' % self.factory.filename)
            dhnio.DprintException()
        if self.factory.do_status_report:
            _SendStatusFunc(self.transport.getPeer(), self.factory.filename, 'failed', 'tcp', err, 'transfer failed')
        if self.factory.result_defer is not None:
            if not self.factory.result_defer.called:
                self.factory.result_defer.errback(Exception('failed'))
                self.factory.result_defer = None
        # dhnio.Dprint(14, 'transport_tcp.SingleSendingProtocol.transferFailed NETERROR host=%s file=%s error=%s' % (str(self.factory.host), str(self.factory.filename), str(err.getErrorMessage())))

    def connectionLost(self, reason):
        global _UnRegisterTransferFunc
        global _ByTransferID
        if self.transfer_id is not None:
            _UnRegisterTransferFunc(self.transfer_id)
            _ByTransferID.pop(self.transfer_id)
        self.transport.loseConnection()
        try:
            self.fin.close()
        except:
            dhnio.DprintException()
        # dhnio.Dprint(14, 'transport_tcp.SingleSendingProtocol.connectionLost with %s' % str(self.transport.getPeer()))

class SingleSendingFactory(protocol.ClientFactory):
    def __init__(self, filename, host, port, result_defer=None, do_status_report=True, send_control_func=None, description=''):
        self.filename = filename
        self.host = host
        self.port = port
        self.protocol = SingleSendingProtocol
        self.result_defer = result_defer
        self.do_status_report = do_status_report
        self.send_control_func = send_control_func # callback which reads from the file
        self.description = description 

    def clientConnectionFailed(self, connector, reason):
        global _SendStatusFunc
        protocol.ClientFactory.clientConnectionFailed(self, connector, reason)
        if self.do_status_report:
            _SendStatusFunc(connector.getDestination(),
                self.filename, 'failed', 'tcp', reason, 'connection failed')
        if self.result_defer is not None:
            if not self.result_defer.called:
                self.result_defer.errback(Exception('failed'))
                self.result_defer = None
        name = str(reason.type.__name__)
        # dhnio.Dprint(14, 'transport_tcp.SingleSendingFactory.clientConnectionFailed NETERROR [%s] with %s:%s' % (
        #     name, connector.getDestination().host, connector.getDestination().port,))

#------------------------------------------------------------------------------

#class SingleReceiveProtocol(protocol.Protocol):
#    def getReceivedBytes(self):
#        return self.receivedBytes
#    
#    def connectionMade(self):
#        global _RegisterTransferFunc
#        global _ByTransferID
#        self.filename = None     # string with path/filename
#        self.fd = None           # integer file descriptor like os.open() returns
#        self.peer = None
#        self.receivedBytes = 0
#        self.transfer_id = None
#        self.blocked = False
#        self.resume_receiving_task = None
#        if self.peer is None:
#            self.peer = self.transport.getPeer()
#        else:
#            if self.peer != self.transport.getPeer():
#                raise Exception("transport_tcp.SingleReceiveProtocol.connectionMade NETERROR thought we had one object per connection")
#        try:
#            from transport_control import black_IPs_dict
#            if self.peer.host in black_IPs_dict().keys() or '.'.join(self.peer.host.split('.')[:-1]) in black_IPs_dict().keys():
#                # dhnio.Dprint(12, 'transport_tcp.SingleReceiveProtocol.connectionMade from BLACK IP: %s, %s' % (str(self.peer.host), str(black_IPs_dict()[self.peer.host])) )
#                self.blocked = True
#                self.transport.loseConnection()
#                return
#        except:
#            pass
#        if self.filename is None:
#            self.fd, self.filename = tmpfile.make("tcp-in")
#        else:   
#            raise Exception("transport_tcp.SingleReceiveProtocol.connectionMade has second connection in same object")
#        if self.fd is None:
#            raise Exception("transport_tcp.SingleReceiveProtocol.connectionMade error opening temporary file")
#        self.transfer_id = _RegisterTransferFunc(
#            'receive', (self.transport.getPeer().host, int(self.transport.getPeer().port)), 
#            self.getReceivedBytes, self.filename, -1)
#        _ByTransferID[self.transfer_id] = self
#
#    def dataReceived(self, data):
#        if self.blocked:
#            return
#        if self.fd is None:
#            raise Exception('transport_tcp.SingleReceiveProtocol.dataReceived from %s but file was not opened' % str(self.peer))
#        amount = len(data)
#        os.write(self.fd, data)
#        self.receivedBytes += amount
#        if self.factory.receive_control_func is not None:
#            seconds_pause = self.factory.receive_control_func(len(data))
#            if seconds_pause > 0:
#                self.transport.pauseProducing()
#                self.resume_receiving_task = reactor.callLater(seconds_pause, self.resumeReceiving)
#
#    def resumeReceiving(self):
#        self.resume_receiving_task = None
#        self.transport.resumeProducing()
#
#    def connectionLost(self, reason):
#        global _ReceiveStatusFunc
#        global _UnRegisterTransferFunc
#        global _ByTransferID
#        if self.resume_receiving_task:
#            self.resume_receiving_task.cancel()
#        if self.transfer_id is not None:
#            _UnRegisterTransferFunc(self.transfer_id)
#            _ByTransferID.pop(self.transfer_id)
#        # dhnio.Dprint(6, 'transport_tcp.SingleReceiveProtocol.connectionLost with %s' % str(self.transport.getPeer()))
#        if self.fd is not None:
#            try:
#                os.close(self.fd)
#            except:
#                dhnio.DprintException()
#        if self.filename is not None:
#            _ReceiveStatusFunc(self.filename, "finished", 'tcp', self.transport.getPeer(), reason)


#class SingleReceiveFactory(protocol.ServerFactory):
#    protocol = SingleReceiveProtocol
#    
#    def __init__(self, receive_control_func):
#        self.receive_control_func = receive_control_func
        
#    def buildProtocol(self, addr):
#        p = SingleReceiveProtocol()
#        p.factory = self
#        return p

#------------------------------------------------------------------------------

class TCPFileSender(basic.FileSender):
    def __init__(self, protocol, inputfile, filename, sz, do_status_report, result_defer, description):
        global _RegisterTransferFunc
        global _ByTransferID
        self.protocol = protocol
        self.inputfile = inputfile
        self.filename = filename
        self.sz = sz
        self.do_status_report = do_status_report 
        self.result_defer = result_defer
        self.description = description
        self.sentBytes = 0
        self.peer = self.protocol.remoteaddress # self.protocol.transport.getPeer()
        self.transfer_id = _RegisterTransferFunc(
            'send', self.peer, self.getSentBytes, filename, sz, description)
        _ByTransferID[self.transfer_id] = self.protocol
        # dhnio.Dprint(14, 'transport_tcp.TCPFileSender.init length=%d transfer_id=%s' % (self.sz, self.transfer_id))

    # def __del__(self):
    #     dhnio.Dprint(14, 'transport_tcp.TCPFileSender.del length=%d transfer_id=%s' % (self.sz, self.transfer_id))

    def getSentBytes(self):
        return self.sentBytes
    
    def transformData(self, data):
        datalength = len(data)
        self.sentBytes += datalength
        self.protocol.totalBytesSent += datalength
        return data

class TCPFileReceiver():
    def __init__(self, protocol, length):
        global _RegisterTransferFunc
        global _ByTransferID
        self.protocol = protocol
        self.receivedBytes = 0
        self.length = length
        self.fd, self.filename = tmpfile.make("tcp-in")
        self.peer = self.protocol.remoteaddress # self.protocol.transport.getPeer()
        self.transfer_id = _RegisterTransferFunc(
            'receive', self.peer, self.getReceivedBytes, self.filename, -1)
        _ByTransferID[self.transfer_id] = self.protocol
        # dhnio.Dprint(14, 'transport_tcp.TCPFileReceiver.init length=%d transfer_id=%s' % (self.length, self.transfer_id))
        
    # def __del__(self):
    #     dhnio.Dprint(14, 'transport_tcp.TCPFileReceiver.del length=%d transfer_id=%s' % (self.length, self.transfer_id))

    def getReceivedBytes(self):
        return self.receivedBytes

    def dataReceived(self, data):
        amount = len(data)
        amt = amount
        tail = ''
        if self.length >= 0:
            moreneeded = self.length - self.receivedBytes
            if amount > moreneeded:
                tail = data[moreneeded:]
                data = data[:moreneeded]
                amount = moreneeded
        self.receivedBytes += amount
        os.write(self.fd, data)
        return tail
        
    def close(self, result, reason):    
        global _UnRegisterTransferFunc
        global _ByTransferID
        global _ReceiveStatusFunc
        if self.transfer_id is not None:
            _UnRegisterTransferFunc(self.transfer_id)
            _ByTransferID.pop(self.transfer_id)
        try:
            os.close(self.fd)
        except:
            dhnio.DprintException()
        _ReceiveStatusFunc(self.filename, result, 'tcp', self.peer, reason)
        
    def eof(self):
        self.close('finished', 'file received successfully')    
        
    def cancel(self):
        self.close('failed', 'transfer cancelled')
        
    def lost(self, result, reason):
        self.close(result, reason)
    

class TCPProtocol(protocol.Protocol):
    def __init__(self):
        self.outbox = []
        self.singleconnection = False
        self.remoteaddress = None
        self.peeraddress = None
        self.remoteinternalport = None
        self.fileSender = None
        self.fileReceiver = None
        self.type = None
        self.totalBytesReceived = 0
        self.totalBytesSent = 0
        self.connected = 0
        self.disconnectingTask = None
    
    def __repr__(self):
        return 'TCPProtocol%s [%d files]' % (self.peeraddress, len(self.outbox))
    
    def connectionMade(self):
        dhnio.Dprint(14, 'transport_tcp.TCPProtocol.connectionMade %s' % str(self.transport.getPeer()))
        global _InternalPort
        global _StartedConnections
        global _OpenedConnections
        global _ConnectionsCounter
        self.connected = time.time()
        self.peeraddress = (self.transport.getPeer().host, int(self.transport.getPeer().port))
        self.remoteaddress = self.peeraddress
        started = _StartedConnections.pop(self.peeraddress, None)
        if started:
            # connecting
            self.remoteinternalport = self.remoteaddress[1]
            self.transport.write('0 port %d ' % _InternalPort)
            self.type = 'client'
        else:
            # listening
            self.type = 'server'
        if self.remoteaddress not in _OpenedConnections:
            _OpenedConnections[self.remoteaddress] = []
        _OpenedConnections[self.remoteaddress].append(self)
        _ConnectionsCounter += 1
        for filename, do_status_report, result_defer, description in self.factory.pendingoutboxfiles:
            self.sendfile(filename, do_status_report, result_defer, description)
        self.factory.pendingoutboxfiles = []
        # dhnio.Dprint(16, '    %s' % str(_OpenedConnections))
        # dhnio.Dprint(16, '    %s' % str(_StartedConnections))
     
    def connectionLost(self, reason):
        global _OpenedConnections
        global _ConnectionsCounter
        global _RedirectionsMap
        dhnio.Dprint(14, 'transport_tcp.TCPProtocol.connectionLost with %s' % str(self.transport.getPeer()))
        self.shutdown(reason)
        if self.remoteaddress in _OpenedConnections:
            try:
                _OpenedConnections[self.remoteaddress].remove(self)
            except:
                pass
            if len(_OpenedConnections[self.remoteaddress]) == 0:
                _OpenedConnections.pop(self.remoteaddress)
#        if self.remoteaddress in _RedirectionsMap:
#            _RedirectionsMap.pop(self.remoteaddress)
#        redirects = _RedirectionsMap.keys()
#        for oldaddress in redirects:
#            if _RedirectionsMap[oldaddress] == self.remoteaddress:
#                _RedirectionsMap.pop(oldaddress)
        _ConnectionsCounter -= 1
        # dhnio.Dprint(16, '    %s' % str(_OpenedConnections))
        # dhnio.Dprint(16, '    %s' % str(_StartedConnections))

    def disconnect(self):
        dhnio.Dprint(14, 'transport_tcp.TCPProtocol.disconnect with %s' % str(self.remoteaddress))
        if self.fileSender is not None:
            self.fileSender.file = None
            self.fileSender.stopProducing()
        try:
            self.transport.abortConnection()
        except:
            self.transport.loseConnection()
    
    def dataReceived(self, data):
        global _OpenedConnections
        global _RedirectionsMap
        if self.remoteinternalport is None:
            try:
                _data = data
                ver, x, _data = _data.partition(' ')
                cmd, x, _data = _data.partition(' ')
                port, x, _data = _data.partition(' ')
                int(ver)
                assert cmd in ['port',]
                self.remoteinternalport = int(port)
                data = _data
            except:
                self.remoteinternalport = self.remoteaddress[1]
                self.singleconnection = True
                # dhnio.DprintException()
                # self.disconnect()
            if self.remoteaddress[1] != self.remoteinternalport:
                oldport = self.remoteaddress[1]
                _OpenedConnections[self.remoteaddress].remove(self)
                if len(_OpenedConnections[self.remoteaddress]) == 0:
                    _OpenedConnections.pop(self.remoteaddress)
                old = self.remoteaddress
                self.remoteaddress = (self.transport.getPeer().host, self.remoteinternalport)
                if self.remoteaddress not in _OpenedConnections:
                    _OpenedConnections[self.remoteaddress] = []
                _OpenedConnections[self.remoteaddress].append(self)
                # _RedirectionsMap[old] = self.remoteaddress
                dhnio.Dprint(12, 'transport_tcp.TCPProtocol.dataReceived %s internal port changed %d->%d' % (
                    self, oldport, self.remoteinternalport))
        if not data:
            return
        if self.fileReceiver is None:
            try:
                _data = data
                ver, x, _data = _data.partition(' ')
                cmd, x, _data = _data.partition(' ')
                length, x, _data = _data.partition(' ')
                int(ver)
                assert cmd in ['length',]
                length = int(length)
                data = _data
            except:
                length = -1
                self.singleconnection = True
                # dhnio.DprintException()
                # self.disconnect()
                # return
            self.fileReceiver = TCPFileReceiver(self, length)
        tail = self.fileReceiver.dataReceived(data)
        self.totalBytesReceived += len(data) - len(tail)
        if self.fileReceiver.length == self.fileReceiver.receivedBytes:
            self.fileReceiver.eof()
            del self.fileReceiver
            self.fileReceiver = None 
            if self.remoteaddress in _OpenedConnections:
                if len(_OpenedConnections[self.remoteaddress]) > 1:
                    if self in _OpenedConnections[self.remoteaddress]:
                        dhnio.Dprint(14, 'transport_tcp.TCPProtocol.finishedTransfer want to close extra connection to %s' % (str(self.remoteaddress)))
                        self.disconnect()
                        return
            if opened_connections_count() > 200:
                dhnio.Dprint(12, 'transport_tcp.TCPProtocol.dataReceived too many opened connections, disconnect with %s' % (str(self.remoteaddress)))
                self.disconnect()
                return
        if tail:
            self.dataReceived(tail)
                    

    def sendfile(self, filename, do_status_report, result_defer, description):
        global _ByTransferID
        global _SendStatusFunc
        # global _SingleConnectionHosts
        if self.type == 'client':
            if self.totalBytesReceived == 0:
                if time.time() - self.connected > 60.0:
                    # _SingleConnectionHosts.add(self.remoteaddress[0])
                    if do_status_report:
                        reactor.callLater(0, _SendStatusFunc, self.remoteaddress, filename, 'failed', 'tcp', None, 'remote peer not responding')
                    if result_defer is not None:
                        if not result_defer.called:
                            result_defer.errback(Exception('failed'))
                    dhnio.Dprint(14, 'transport_tcp.TCPProtocol.sendfile connection idle, disconnect with %s' % (str(self.remoteaddress)))        
                    self.disconnect()
                    return
        if not os.path.isfile(filename):
            if do_status_report:
                reactor.callLater(0, _SendStatusFunc, self.remoteaddress, filename, 'failed', 'tcp', None, 'file not exist')
            if result_defer is not None:
                if not result_defer.called:
                    result_defer.errback(Exception('failed'))
            dhnio.Dprint(14, 'transport_tcp.TCPProtocol.sendfile sending file not exist, disconnect with %s' % (str(self.remoteaddress)))
            self.disconnect()
            return
        try:
            sz = os.path.getsize(filename)
        except:
            dhnio.DprintException()
            if do_status_report:
                reactor.callLater(0, _SendStatusFunc, self.remoteaddress, filename, 'failed', 'tcp', None, 'error opening file')
            if result_defer is not None:
                if not result_defer.called:
                    result_defer.errback(Exception('failed'))
            dhnio.Dprint(14, 'transport_tcp.TCPProtocol.sendfile reading file size ERROR, disconnect with %s' % (str(self.remoteaddress)))
            self.disconnect()
            return
        self.outbox.append((filename, sz, do_status_report, result_defer, description))  
        reactor.callLater(0, self.processoutbox)
        
    def processoutbox(self):
        global _SendStatusFunc
        if len(self.outbox) == 0:
            return
        if self.fileSender is not None:
            return
        filename, sz, do_status_report, result_defer, description = self.outbox.pop(0)
        try:
            fin = open(filename, 'rb')
        except:
            dhnio.DprintException()
            if do_status_report:
                _SendStatusFunc(self.remoteaddress, filename, 'failed', 'tcp', None, 'error opening file')
            if result_defer is not None:
                if not result_defer.called:
                    result_defer.errback(Exception('failed'))
            dhnio.Dprint(14, 'transport_tcp.TCPProtocol.sendfile opening file ERROR, disconnect with %s' % (str(self.remoteaddress)))
            self.disconnect()
            return
        self.transport.write('0 length %d ' % sz)
        self.fileSender = TCPFileSender(self, fin, filename, sz, do_status_report, result_defer, description)
        d = self.fileSender.beginFileTransfer(fin, self.transport, self.fileSender.transformData)
        d.addCallback(self.finishedTransfer)
        d.addErrback(self.transferFailed)

    def finishedTransfer(self, result):
        global _UnRegisterTransferFunc
        global _ByTransferID
        global _SendStatusFunc
        global _OpenedConnections
        # global _SingleConnectionHosts
        # dhnio.Dprint(14, 'transport_tcp.TCPProtocol.finishedTransfer host=%s file=%s' % (
        #     str((self.transport.getPeer().host, int(self.transport.getPeer().port))), self.fileSender.filename,))
        try:
            self.fileSender.inputfile.close()
        except:
            dhnio.Dprint(1, 'transport_tcp.TCPProtocol.finishedTransfer ERROR close file failed')
            dhnio.DprintException()
        if self.fileSender.do_status_report:
            _SendStatusFunc(self.remoteaddress, self.fileSender.filename, 'finished', 'tcp')
        if self.fileSender.result_defer is not None:
            if not self.fileSender.result_defer.called:
                self.fileSender.result_defer.callback('finished')
        if self.fileSender.transfer_id is not None:
            _UnRegisterTransferFunc(self.fileSender.transfer_id)
            _ByTransferID.pop(self.fileSender.transfer_id)
        self.fileSender.transform = None
        del self.fileSender
        self.fileSender = None
        if opened_connections_count() > 200:
            dhnio.Dprint(12, 'transport_tcp.TCPProtocol.finishedTransfer too many opened connections, disconnect with %s' % (str(self.remoteaddress)))
            self.disconnect()
            return
        if self.singleconnection or self.remoteinternalport is None:
            dhnio.Dprint(14, 'transport_tcp.TCPProtocol.finishedTransfer singleconnection=%s remoteinternalport=%s, disconnect with %s' % (
                self.singleconnection, self.remoteinternalport, str(self.remoteaddress)))
            self.transport.disconnect()
            # _SingleConnectionHosts.add(self.remoteaddress[0])
        else:
            reactor.callLater(0, self.processoutbox)

    def transferFailed(self, err):
        global _SendStatusFunc
        global _UnRegisterTransferFunc
        global _ByTransferID
        global _OpenedConnections
        # global _SingleConnectionHosts
        # dhnio.Dprint(14, 'transport_tcp.TCPProtocol.transferFailed NETERROR host=%s file=%s error=%s' % (
        #     str((self.transport.getPeer().host, int(self.transport.getPeer().port))), self.fileSender.filename, str(err.getErrorMessage())))
        try:
            self.fileSender.inputfile.close()
        except:
            dhnio.Dprint (1, 'transport_tcp.TCPProtocol.transferFailed ERROR closing file %s' % self.fileSender.filename)
            dhnio.DprintException()
        if self.fileSender.do_status_report:
            _SendStatusFunc(self.remoteaddress, self.fileSender.filename, 'failed', 'tcp', err, 'transfer failed')
        if self.fileSender.result_defer is not None:
            if not self.fileSender.result_defer.called:
                self.fileSender.result_defer.errback(Exception('failed'))
        if self.fileSender.transfer_id is not None:
            _UnRegisterTransferFunc(self.fileSender.transfer_id)
            _ByTransferID.pop(self.fileSender.transfer_id, None)
        self.fileSender.transform = None
        del self.fileSender
        self.fileSender = None
        if opened_connections_count() > 200:
            dhnio.Dprint(12, 'transport_tcp.TCPProtocol.finishedTransfer too many opened connections, disconnect with %s' % (str(self.remoteaddress)))
            self.disconnect()
            return
        if self.singleconnection or self.remoteinternalport is None:
            dhnio.Dprint(14, 'transport_tcp.TCPProtocol.transferFailed singleconnection=%s remoteinternalport=%s, disconnect with %s' % (
                self.singleconnection, self.remoteinternalport, str(self.remoteaddress)))
            self.transport.disconnect()
            # _SingleConnectionHosts.add(self.remoteaddress[0])
        else:
            reactor.callLater(0, self.processoutbox)

    def shutdown(self, reason=None):
        global _SendStatusFunc
        global _UnRegisterTransferFunc
        global _ByTransferID
        if self.fileReceiver is not None:
            typ = getattr(reason, 'type', None)
            if str(typ).count('ConnectionDone'): 
                self.fileReceiver.lost('finished', reason)
            else:
                self.fileReceiver.lost('failed', reason)
            del self.fileReceiver
            self.fileReceiver = None
        if self.fileSender is not None:
            try:
                self.fileSender.inputfile.close()
            except:
                dhnio.DprintException()
            self.fileSender.file = None
            if self.fileSender.do_status_report:
                _SendStatusFunc(self.remoteaddress, self.fileSender.filename, 'failed', 'tcp', None, 'connection lost')
            if self.fileSender.result_defer is not None:
                if not self.fileSender.result_defer.called:
                    self.fileSender.result_defer.errback(Exception('failed'))
            if self.fileSender.transfer_id is not None:
                _UnRegisterTransferFunc(self.fileSender.transfer_id)
                _ByTransferID.pop(self.fileSender.transfer_id, None)
            self.transport.unregisterProducer()
            del self.fileSender
            self.fileSender = None
        pending_transfer_ids = []
        for transfer_id, proto in _ByTransferID.items():
            if proto == self:
                pending_transfer_ids.append(transfer_id)
        for transfer_id in pending_transfer_ids:
            _UnRegisterTransferFunc(transfer_id)
            _ByTransferID.pop(transfer_id, None)

    
class TCPFactory(protocol.ClientFactory):
    protocol = TCPProtocol
    
    def __init__(self):
        self.pendingoutboxfiles = []

    def __repr__(self):
        return 'TCPFactory' 

    def clientConnectionFailed(self, connector, reason):
        global _SendStatusFunc
        global _StartedConnections
        protocol.ClientFactory.clientConnectionFailed(self, connector, reason)
        destaddress = (connector.getDestination().host, int(connector.getDestination().port))
        for filename, do_status_report, result_defer, description in self.pendingoutboxfiles:
            if do_status_report:
                _SendStatusFunc(destaddress, filename, 'failed', 'tcp', reason, 'connection failed')
            if result_defer is not None:
                if not result_defer.called:
                    result_defer.errback(Exception('failed'))
                    result_defer = None
        # dhnio.Dprint(14, 'transport_tcp.TCPProtocol.clientConnectionFailed opened NETERROR [%s] with %s:%s, id=%d' % (
        #     str(reason.type.__name__), connector.getDestination().host, connector.getDestination().port, id(self.pendingoutboxfiles)))
        _StartedConnections.pop(destaddress, None)
        
    def addoutboxfile(self, filename, do_status_report, result_defer, description):
        self.pendingoutboxfiles.append((filename, do_status_report, result_defer, description))

#------------------------------------------------------------------------------ 

def SendStatusFuncDefault(host, filename, status, proto='', error=None, message=''):
    try:
        from transport_control import sendStatusReport
        sendStatusReport(host, filename, status, proto, error, message)
    except:
        dhnio.DprintException()

def ReceiveStatusFuncDefault(filename, status, proto='', host=None, error=None, message=''):
    try:
        from transport_control import receiveStatusReport
        receiveStatusReport(filename, status, proto, host, error, message)
    except:
        dhnio.DprintException()

def SendControlFuncDefault(prev_read, chunk_size):
    return chunk_size

def ReceiveControlFuncDefault(new_data_size):
    return 0

def RegisterTransferFuncDefault(send_or_receive=None, remote_address=None, callback=None, filename=None, size=None, description='', transfer_id=None):
    try:
        from transport_control import register_transfer
        return register_transfer('tcp', send_or_receive, remote_address, callback, filename, size, description, transfer_id)
    except:
        dhnio.DprintException()
        return None
    
def UnRegisterTransferFuncDefault(transfer_id):
    try:
        from transport_control import unregister_transfer
        return unregister_transfer(transfer_id)
    except:
        dhnio.DprintException()
    return None

#------------------------------------------------------------------------------ 

_SendStatusFunc = SendStatusFuncDefault
_ReceiveStatusFunc = ReceiveStatusFuncDefault
_SendControlFunc = SendControlFuncDefault
_ReceiveControlFunc = ReceiveControlFuncDefault
_RegisterTransferFunc = RegisterTransferFuncDefault
_UnRegisterTransferFunc = UnRegisterTransferFuncDefault

#-------------------------------------------------------------------------------

def usage():
        print '''
args:
transport_tcp.py receive [port]                                        to receive
transport_tcp.py send [host] [port] [file name]                        to send a file
transport_tcp.py send [host] [port] [file name] [interval]             to start sending continuously
transport_tcp.py opensend [myport] [peerhost] [peerport] [file name]   send and keep connection opened
'''



#def mytest():
#    dhnio.SetDebug(18)
#
#    filename = "transport_tcp.pyc"              # just some file to send
###    host = "work.offshore.ai"
###    host = "89.223.30.208"
#    host = 'localhost'
#    port = 7771
#
#    if len(sys.argv) == 4:
#        send(sys.argv[3], sys.argv[1], sys.argv[2])
#    elif len(sys.argv) == 2:
#        r = receive(sys.argv[1])
#        print r
#    else:
#        usage()
#        sys.exit()
#    reactor.run()

#bytes_counter = 0
#time_counter = time.time()
#count_period = 1
#current_chunk = 1

def _my_receive_control(new_data_size):
    try:
        from transport_control import ReceiveTrafficControl
        return ReceiveTrafficControl(new_data_size)
    except:
        dhnio.DprintException()
        return 0
    
def _my_send_control(prev_read_size, chunk_size):
    try:
        from transport_control import SendTrafficControl
        return SendTrafficControl(prev_read_size, chunk_size)
    except:
        dhnio.DprintException()
        return chunk_size

def _my_monitor():
#    import transport_control
#    src = ''
#    for address in transport_control._LiveTransfers.keys():
#        for obj in transport_control._LiveTransfers[address]['send']:
#            src += '%d to %s, ' % (obj.sentBytes, str(address))
#        for obj in transport_control._LiveTransfers[address]['receive']:
#            src += '%d from %s, ' % (obj.receivedBytes, str(address))
#    print '>>>', src
    print _OpenedConnections, _StartedConnections
    reactor.callLater(5, _my_monitor)

if __name__ == "__main__":
    #import datahaven.p2p.memdebug as memdebug
    #memdebug.start(8080)
    dhnio.init()
    dhnio.SetDebug(16)
    dhnio.LifeBegins()

    from twisted.internet.defer import setDebugging
    setDebugging(True)

    _my_monitor()

    def _test_cancel():
        import transport_control as tc
        for t in tc.current_transfers():
            cancel(t.transfer_id)

    if len(sys.argv) < 2:
        usage()
        sys.exit(1)
    
    cmd = sys.argv[1].lower().strip()

    if cmd == 'receive':
        r = receive(sys.argv[2])
        reactor.callLater(5, disconnect, '127.0.0.1', 6661)
        reactor.run()
    elif cmd == 'send':
        _SendStatusFunc = lambda host, filename, status, proto='', error=None, message='': \
            sys.stdout.write('                        %s to %s is %s\n' % (filename, host, status))
        if len(sys.argv) < 6:
            send(sys.argv[4], sys.argv[2], sys.argv[3]) # .addBoth(lambda x: dhnio.Dprint(2, 'RESULT:%s'%str(x)))
            # reactor.callLater(10, _test_cancel)
        else:
            l = task.LoopingCall(send, sys.argv[4], sys.argv[2], sys.argv[3])
            l.start(float(sys.argv[5]))
            reactor.callLater(10, _test_cancel)
            reactor.callLater(20, _test_cancel)
        reactor.run()
    elif cmd == 'opensend':
        r = receive(sys.argv[2])
        _SendStatusFunc = lambda host, filename, status, proto='', error=None, message='': \
            sys.stdout.write('                        %s to %s is %s\n' % (filename, host, status))
        reactor.callLater(1, send, sys.argv[5], sys.argv[3], sys.argv[4]) # .addBoth(lambda x: dhnio.Dprint(2, 'RESULT:%s'%str(x)))
        # reactor.callLater(5, disconnect, sys.argv[3], sys.argv[4])
        reactor.run()
    else:
        usage()

